#%%
"""
Reference:
[1] Synthcity: facilitating innovative use cases of synthetic data in different data modalities
- https://github.com/vanderschaarlab/synthcity/blob/main/src/synthcity/metrics/eval_privacy.py
"""
#%%
import numpy as np
import pandas as pd
from collections import Counter

from sklearn import metrics
from sklearn.cluster import KMeans
from sklearn.preprocessing import StandardScaler, MinMaxScaler, OneHotEncoder
from scipy.spatial import distance_matrix
#%%
def kAnonymization(train, syndata, continuous_features):
    """
    Returns the k-anonimity ratio between the real data and the synthetic data.
    For each dataset, it is computed the value k which satisfies the k-anonymity rule: 
    each record is similar to at least another k-1 other records on the potentially identifying variables.
    
    For comparison between different datasets, we normalize using the number of observations.
    : higher is better
    """
    
    train = train[continuous_features]
    syndata = syndata[continuous_features]
    
    scaler = StandardScaler().fit(train)
    train = scaler.transform(train)
    syndata = scaler.transform(syndata)
    
    def evaluate_data(data):
        values = [999]
        for n_clusters in [2, 5, 10, 15]:
            if len(data) / n_clusters < 10:
                continue
            cluster = KMeans(
                n_clusters=n_clusters, init="k-means++", random_state=0
            ).fit(data)
            counts: dict = Counter(cluster.labels_)
            values.append(np.min(list(counts.values())))
        return np.min(values) / len(data) * 100
    
    return (
        evaluate_data(train),
        evaluate_data(syndata)
    )
#%%
def kMap(train, syndata, continuous_features):
    """
    Returns the minimum value k that satisfies the k-map rule.
    The data satisfies k-map if every combination of values for the quasi-identifiers appears at least k times in the reidentification(synthetic) dataset.
    
    For comparison between different datasets, we normalize using the number of observations.
    : higher is better
    """
    
    train = train[continuous_features]
    syndata = syndata[continuous_features]
    
    scaler = StandardScaler().fit(train)
    train = scaler.transform(train)
    syndata = scaler.transform(syndata)
    
    values = []
    for n_clusters in [2, 5, 10, 15]:
        if len(train) / n_clusters < 10:
            continue
        cluster = KMeans(
            n_clusters=n_clusters, init="k-means++", random_state=0
        ).fit(train)
        clusters = cluster.predict(syndata)
        counts: dict = Counter(clusters)
        values.append(np.min(list(counts.values())))

    if len(values) == 0:
        return 0

    return np.min(values) / len(syndata) * 100
#%%
def DCR_metric(train, syndata, continuous_features, categorical_features, data_percent=15):
    """
    Reference:
    [1] https://github.com/Team-TUD/CTAB-GAN/blob/main/model/eval/evaluation.py
    
    Returns Distance to Closest Record
    
    Inputs:
    1) train -> real data
    2) synthetic -> corresponding synthetic data
    3) data_percent -> percentage of data to be sampled from real and synthetic datasets for computing Distance to Closest Record
    Outputs:
    1) List containing the 5th percentile distance to closest record (DCR) between real and synthetic as well as within real and synthetic datasets
    along with 5th percentile of nearest neighbour distance ratio (NNDR) between real and synthetic as well as within real and synthetic datasets
    
    : higher is better
    """
    
    ### pre-processing
    train_ = train.copy()
    syndata_ = syndata.copy()
    # continuous: min-max scaling
    scaler = MinMaxScaler().fit(train_[continuous_features])
    train_[continuous_features] = scaler.transform(train_[continuous_features])
    syndata_[continuous_features] = scaler.transform(syndata_[continuous_features])
    # categorical: one-hot encoding
    scaler = OneHotEncoder(handle_unknown='ignore').fit(train_[categorical_features])
    train_ = np.concatenate([
        train_[continuous_features].values,
        scaler.transform(train_[categorical_features]).toarray()
    ], axis=1)
    syndata_ = np.concatenate([
        syndata_[continuous_features].values,
        scaler.transform(syndata_[categorical_features]).toarray()
    ], axis=1)
    
    train_ = pd.DataFrame(train_)
    syndata_ = pd.DataFrame(syndata_)
    
    # Sampling smaller sets of real and synthetic data to reduce the time complexity of the evaluation
    real_sampled = train_.sample(n=int(len(train_)*(.01*data_percent)), random_state=42).to_numpy()
    fake_sampled = syndata_.sample(n=int(len(syndata_)*(.01*data_percent)), random_state=42).to_numpy()

    # Computing pair-wise distances between real and synthetic 
    dist_rf = metrics.pairwise_distances(real_sampled, Y=fake_sampled, metric='minkowski', n_jobs=-1)
    # Computing pair-wise distances within real 
    dist_rr = metrics.pairwise_distances(real_sampled, Y=None, metric='minkowski', n_jobs=-1)
    # Computing pair-wise distances within synthetic
    dist_ff = metrics.pairwise_distances(fake_sampled, Y=None, metric='minkowski', n_jobs=-1) 
    
    # Removes distances of data points to themselves to avoid 0s within real and synthetic 
    rd_dist_rr = dist_rr[~np.eye(dist_rr.shape[0],dtype=bool)].reshape(dist_rr.shape[0],-1)
    rd_dist_ff = dist_ff[~np.eye(dist_ff.shape[0],dtype=bool)].reshape(dist_ff.shape[0],-1) 
    
    # Computing first and second smallest nearest neighbour distances between real and synthetic
    smallest_two_indexes_rf = [dist_rf[i].argsort()[:2] for i in range(len(dist_rf))]
    smallest_two_rf = [dist_rf[i][smallest_two_indexes_rf[i]] for i in range(len(dist_rf))]       
    # Computing first and second smallest nearest neighbour distances within real
    smallest_two_indexes_rr = [rd_dist_rr[i].argsort()[:2] for i in range(len(rd_dist_rr))]
    smallest_two_rr = [rd_dist_rr[i][smallest_two_indexes_rr[i]] for i in range(len(rd_dist_rr))]
    # Computing first and second smallest nearest neighbour distances within synthetic
    smallest_two_indexes_ff = [rd_dist_ff[i].argsort()[:2] for i in range(len(rd_dist_ff))]
    smallest_two_ff = [rd_dist_ff[i][smallest_two_indexes_ff[i]] for i in range(len(rd_dist_ff))]
    
    # Computing 5th percentiles for DCR and NNDR between and within real and synthetic datasets
    min_dist_rf = np.array([i[0] for i in smallest_two_rf])
    fifth_perc_rf = np.percentile(min_dist_rf,5)
    min_dist_rr = np.array([i[0] for i in smallest_two_rr])
    fifth_perc_rr = np.percentile(min_dist_rr,5)
    min_dist_ff = np.array([i[0] for i in smallest_two_ff])
    fifth_perc_ff = np.percentile(min_dist_ff,5)
    
    return [fifth_perc_rf, fifth_perc_rr, fifth_perc_ff]
#%%
def attribute_disclosure(K, compromised, syndata, attr_compromised, continuous_features, categorical_features):
    dist = distance_matrix(
        compromised[attr_compromised].to_numpy(),
        syndata[attr_compromised].to_numpy(),
        p=2)
    K_idx = dist.argsort(axis=1)[:, :K]
    
    def most_common(lst):
        return max(set(lst), key=lst.count)
    
    votes = []
    trues = []
    for i in range(len(K_idx)):
        true = np.zeros((len(categorical_features), ))
        vote = np.zeros((len(categorical_features), ))
        for j in range(len(categorical_features)):
            true[j] = compromised.to_numpy()[i, len(continuous_features) + j]
            vote[j] = most_common(list(syndata.to_numpy()[K_idx[i], len(continuous_features) + j]))
        votes.append(vote)
        trues.append(true)
    votes = np.vstack(votes)
    trues = np.vstack(trues)
    
    acc = 0
    for j in range(trues.shape[1]):
        acc += (trues[:, j] == votes[:, j]).mean()
    acc /= trues.shape[1]
    return acc

def AttributeDisclosure(train, syndata, continuous_features, categorical_features):
    """
    Reference:
    [1] A Review of Attribute Disclosure Control
    
    : lower is better
    """
    
    train_ = train.copy()
    syndata_ = syndata.copy()
    
    continuous = continuous_features
    scaler = StandardScaler().fit(train_[continuous])
    train_[continuous] = scaler.transform(train_[continuous])
    syndata_[continuous] = scaler.transform(syndata_[continuous])
    
    compromised_idx = np.random.choice(
        range(len(train_)), 
        int(len(train_) * 0.01), 
        replace=False)
    compromised = train_.iloc[compromised_idx].reset_index().drop(columns=['index'])
    
    result = []
    for attr_num in [3, 5, 10, 15]:
        if attr_num > len(continuous_features):
            continue
        else:
            attr_compromised = continuous_features[:attr_num]
            for K in [1, 10, 100]:
                acc = attribute_disclosure(
                    K, compromised, syndata_.copy(), attr_compromised, continuous_features, categorical_features)
                result.append(acc)
    AD = np.mean(result)
    return AD
#%%


#%%
"""
Reference:
[1] Reimagining Synthetic Tabular Data Generation through Data-Centric AI: A Comprehensive Benchmark
- https://github.com/HLasse/data-centric-synthetic-data
"""
#%%
import numpy as np

from sklearn.metrics import accuracy_score
from sklearn.linear_model import LogisticRegression
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.ensemble import AdaBoostClassifier, GradientBoostingClassifier, RandomForestClassifier
from sklearn.preprocessing import StandardScaler, OneHotEncoder
from scipy.stats import spearmanr
#%%
def MLu_cls(train, test, syndata, target, continuous_features, categorical_features):
    ### pre-processing
    continuous = continuous_features
    categorical = [x for x in categorical_features if x != target]
    
    train_ = train.copy()
    test_ = test.copy()
    syndata_ = syndata.copy()
    # continuous: standardization
    scaler = StandardScaler().fit(train_[continuous])
    train_[continuous] = scaler.transform(train_[continuous])
    test_[continuous] = scaler.transform(test_[continuous])
    syndata_[continuous] = scaler.transform(syndata_[continuous])
    # categorical: one-hot encoding
    if len(categorical):
        scaler = OneHotEncoder(drop="first", handle_unknown='ignore').fit(train_[categorical])
        train_ = np.concatenate([
            train_[continuous].values,
            scaler.transform(train_[categorical]).toarray(),
            train_[[target]].values
        ], axis=1)
        test_ = np.concatenate([
            test_[continuous].values,
            scaler.transform(test_[categorical]).toarray(),
            test_[[target]].values
        ], axis=1)
        syndata_ = np.concatenate([
            syndata_[continuous].values,
            scaler.transform(syndata_[categorical]).toarray(),
            syndata_[[target]].values
        ], axis=1)
    else:
        train_ = train_.values
        test_ = test_.values
        syndata_ = syndata_.values
    
    """Baseline"""
    performance = []
    feature_importance = []
    print(f"(Baseline) Classification: Accuracy...")
    for name, clf in [
        ('logit', LogisticRegression(tol=0.001, random_state=42, n_jobs=-1, max_iter=1000)),
        ('KNN', KNeighborsClassifier(n_jobs=-1)),
        ('RBF-SVM', SVC(random_state=42)),
        ('RandomForest', RandomForestClassifier(random_state=42, n_jobs=-1)),
        ('GradBoost', GradientBoostingClassifier(random_state=42)),
        ('AdaBoost', AdaBoostClassifier(random_state=42)),
    ]:
        clf.fit(train_[:, :-1], train_[:, -1])
        pred = clf.predict(test_[:, :-1])
        acc = accuracy_score(test_[:, -1], pred)
        if name in ["RandomForest", "GradBoost", "AdaBoost"]: 
            feature_importance.append(clf.feature_importances_)
        print(f"[{name}] ACC: {acc:.3f}")
        performance.append(acc)

    base_performance = performance
    base_cls = np.mean(performance)
    base_feature_importance = feature_importance
    
    """Synthetic"""
    if len(np.unique(syndata_[:, -1])) == 0:
        return (
            base_cls, 0., 0., 0.
        )
    else:
        performance = []
        feature_importance = []
        print(f"(Synthetic) Classification: Accuracy...")
        for name, clf in [
            ('logit', LogisticRegression(tol=0.001, random_state=42, n_jobs=-1, max_iter=1000)),
            ('KNN', KNeighborsClassifier(n_jobs=-1)),
            ('RBF-SVM', SVC(random_state=42)),
            ('RandomForest', RandomForestClassifier(random_state=42, n_jobs=-1)),
            ('GradBoost', GradientBoostingClassifier(random_state=42)),
            ('AdaBoost', AdaBoostClassifier(random_state=42)),
        ]:
            clf.fit(syndata_[:, :-1], syndata_[:, -1])
            pred = clf.predict(test_[:, :-1])
            acc = accuracy_score(test_[:, -1], pred)
            if name in ["RandomForest", "GradBoost", "AdaBoost"]: 
                feature_importance.append(clf.feature_importances_)
            print(f"[{name}] ACC: {acc:.3f}")
            performance.append(acc)
                
        syn_cls = np.mean(performance)
        model_selection = spearmanr(base_performance, performance).statistic
        feature_selection = []
        for f1, f2 in zip(base_feature_importance, feature_importance):
            feature_selection.append(spearmanr(f1, f2).statistic)
        feature_selection = np.mean(feature_selection)
        
        return (
            base_cls, syn_cls, model_selection, feature_selection
        )
#%%


#%%
import numpy as np
import pandas as pd
#%%
def get_frequency(
    X_gt: pd.DataFrame, X_synth: pd.DataFrame, n_histogram_bins: int = 10
):
    """
    Reference:
    [1] https://github.com/vanderschaarlab/synthcity/blob/main/src/synthcity/metrics/_utils.py
    
    Get percentual frequencies for each possible real categorical value.

    Returns:
        The observed and expected frequencies (as a percent).
    """
    res = {}
    for col in X_gt.columns:
        local_bins = min(n_histogram_bins, len(X_gt[col].unique()))

        if len(X_gt[col].unique()) < 5:  # categorical
            gt = (X_gt[col].value_counts() / len(X_gt)).to_dict()
            synth = (X_synth[col].value_counts() / len(X_synth)).to_dict()
        else:
            gt_vals, bins = np.histogram(X_gt[col], bins=local_bins)
            synth_vals, _ = np.histogram(X_synth[col], bins=bins)
            gt = {k: v / (sum(gt_vals) + 1e-8) for k, v in zip(bins, gt_vals)}
            synth = {k: v / (sum(synth_vals) + 1e-8) for k, v in zip(bins, synth_vals)}

        for val in gt:
            if val not in synth or synth[val] == 0:
                synth[val] = 1e-11
        for val in synth:
            if val not in gt or gt[val] == 0:
                gt[val] = 1e-11

        if gt.keys() != synth.keys():
            raise ValueError(f"Invalid features. {gt.keys()}. syn = {synth.keys()}")
        res[col] = (list(gt.values()), list(synth.values()))

    return res
#%%


#%%
"""
Reference:
[1] Synthcity: facilitating innovative use cases of synthetic data in different data modalities
- https://github.com/vanderschaarlab/synthcity/blob/main/src/synthcity/metrics/eval_statistical.py
"""
#%%
import numpy as np
from tqdm import tqdm
import torch

from scipy.stats import entropy, ks_2samp
from sklearn import metrics
from sklearn.preprocessing import MinMaxScaler, OneHotEncoder
from sklearn.neighbors import NearestNeighbors

from . import utility
#%%
def KLDivergence(train, syndata, continuous_features):
    """
    Marginal statistical fidelity: KL-Divergence
    : lower is better
    """
    
    ### square-root choice
    n = len(train)
    num_bins = int(np.ceil(np.sqrt(n)))
    
    # get distribution of continuous variables
    cont_freqs = utility.get_frequency(
        train[continuous_features], 
        syndata[continuous_features], 
        n_histogram_bins=num_bins
    )
    
    result = [] 
    for col in train.columns:
        if col in continuous_features:
            gt, syn = cont_freqs[col]
            kl_div = entropy(syn, gt)
        else:
            pmf_p = train[col].value_counts(normalize=True)
            pmf_q = syndata[col].value_counts(normalize=True)
            
            # Ensure that both PMFs cover the same set of categories
            all_categories = pmf_p.index.union(pmf_q.index)
            pmf_p = pmf_p.reindex(all_categories, fill_value=0)
            pmf_q = pmf_q.reindex(all_categories, fill_value=0)
            
            # Avoid division by zero and log(0) by filtering out zero probabilities
            non_zero_mask = (pmf_p > 0) & (pmf_q > 0)
            
            kl_div = np.sum(pmf_q[non_zero_mask] * np.log(pmf_q[non_zero_mask] / pmf_p[non_zero_mask]))
        result.append(kl_div)
    return np.mean(result)
#%%
def GoodnessOfFit(train, syndata, continuous_features):
    """
    Marginal statistical fidelity: Kolmogorov-Smirnov test & Chi-Squared test
    : lower is better
    """
    
    result = [] 
    for col in train.columns:
        if col in continuous_features:
            # Compute the Kolmogorov-Smirnov test for goodness of fit.
            statistic, _ = ks_2samp(train[col], syndata[col])
        else:
            pmf_p = train[col].value_counts(normalize=True) # expected
            pmf_q = syndata[col].value_counts(normalize=True) # observed
            
            # Ensure that both PMFs cover the same set of categories
            all_categories = pmf_p.index.union(pmf_q.index)
            pmf_p = pmf_p.reindex(all_categories, fill_value=0)
            pmf_q = pmf_q.reindex(all_categories, fill_value=0)
            
            # Avoid division by zero and log(0) by filtering out zero probabilities
            non_zero_mask = pmf_p > 0
            
            # Compute the Chi-Squared test for goodness of fit.
            statistic = ((pmf_q[non_zero_mask] - pmf_p[non_zero_mask]) ** 2 / pmf_p[non_zero_mask]).sum()
        result.append(statistic)
    return np.mean(result)
#%%
def MaximumMeanDiscrepancy(train, syndata, continuous_features, categorical_features):
    """
    Joint statistical fidelity: Maximum Mean Discrepancy (MMD)
    : lower is better
    
    - MMD using RBF (gaussian) kernel (i.e., k(x,y) = exp(-gamma * ||x-y||^2 / 2))
    """
    
    ### pre-processing
    train_ = train.copy()
    syndata_ = syndata.copy()
    # continuous: min-max scaling
    scaler = MinMaxScaler().fit(train_[continuous_features])
    train_[continuous_features] = scaler.transform(train_[continuous_features])
    syndata_[continuous_features] = scaler.transform(syndata_[continuous_features])
    # categorical: one-hot encoding
    scaler = OneHotEncoder(handle_unknown='ignore').fit(train_[categorical_features])
    train_ = np.concatenate([
        train_[continuous_features].values,
        scaler.transform(train_[categorical_features]).toarray()
    ], axis=1)
    syndata_ = np.concatenate([
        syndata_[continuous_features].values,
        scaler.transform(syndata_[categorical_features]).toarray()
    ], axis=1)
    
    XX = metrics.pairwise.rbf_kernel(
        train_.reshape(len(train_), -1),
        train_.reshape(len(train_), -1),
    )
    YY = metrics.pairwise.rbf_kernel(
        syndata_.reshape(len(syndata_), -1),
        syndata_.reshape(len(syndata_), -1),
    )
    XY = metrics.pairwise.rbf_kernel(
        train_.reshape(len(train_), -1),
        syndata_.reshape(len(syndata_), -1),
    )
    MMD = XX.mean() + YY.mean() - 2 * XY.mean() # squared MMD
    return np.sqrt(MMD)
#%%
def phi(s, D):
    return (1 + (4 * s) / (2 * D - 3)) ** (-1 / 2)

def cramer_wold_distance_function(x_batch, x_gen):
    gamma_ = (4 / (3 * x_batch.size(0))) ** (2 / 5)
    
    cw1 = torch.cdist(x_batch, x_batch) ** 2 
    cw2 = torch.cdist(x_gen, x_gen) ** 2 
    cw3 = torch.cdist(x_batch, x_gen) ** 2 
    cw_x = phi(cw1 / (4 * gamma_), D=x_batch.size(1)).sum()
    cw_x += phi(cw2 / (4 * gamma_), D=x_batch.size(1)).sum()
    cw_x += -2 * phi(cw3 / (4 * gamma_), D=x_batch.size(1)).sum()
    cw_x /= (2 * x_batch.size(0) ** 2 * torch.tensor(torch.pi * gamma_).sqrt())
    return cw_x

def CramerWoldDistance(train, syndata, continuous_features, categorical_features, device):
    """
    Joint statistical fidelity: Cramer-Wold Distance
    : lower is better
    """
    
    ### pre-processing
    train_ = train.copy()
    syndata_ = syndata.copy()
    # continuous: min-max scaling
    scaler = MinMaxScaler().fit(train_[continuous_features])
    train_[continuous_features] = scaler.transform(train_[continuous_features])
    syndata_[continuous_features] = scaler.transform(syndata_[continuous_features])
    # categorical: one-hot encoding
    scaler = OneHotEncoder(handle_unknown='ignore').fit(train_[categorical_features])
    train_ = np.concatenate([
        train_[continuous_features].values,
        scaler.transform(train_[categorical_features]).toarray()
    ], axis=1)
    syndata_ = np.concatenate([
        syndata_[continuous_features].values,
        scaler.transform(syndata_[categorical_features]).toarray()
    ], axis=1)
    train_ = torch.tensor(train_).to(device)
    syndata_ = torch.tensor(syndata_).to(device)
    
    if len(train_) > 10000: # large dataset case
        CWs = []
        for _ in tqdm(range(10), desc="Batch Cramer-Wold Distance..."):
            idx = np.random.choice(range(len(train_)), 2000, replace=False)
            train_small = train_[idx, :]
            syndata_small = syndata_[idx, :]
            cw = cramer_wold_distance_function(train_small, syndata_small)
            CWs.append(cw.cpu().numpy().item())
        cw = np.mean(CWs)
    else:
        cw = cramer_wold_distance_function(train_, syndata_)
        cw = cw.cpu().numpy().item()
    return np.sqrt(cw) # square-root distance
#%%
def naive_alpha_precision_beta_recall(train, syndata, continuous_features, categorical_features):
    """
    Reference:
    - https://github.com/vanderschaarlab/synthcity/blob/main/src/synthcity/metrics/eval_statistical.py
    """
    ### pre-processing
    train_ = train.copy()
    syndata_ = syndata.copy()
    # continuous: min-max scaling
    scaler = MinMaxScaler().fit(train_[continuous_features])
    train_[continuous_features] = scaler.transform(train_[continuous_features])
    syndata_[continuous_features] = scaler.transform(syndata_[continuous_features])
    # categorical: one-hot encoding
    scaler = OneHotEncoder(handle_unknown='ignore').fit(train_[categorical_features])
    train_ = np.concatenate([
        train_[continuous_features].values,
        scaler.transform(train_[categorical_features]).toarray()
    ], axis=1)
    syndata_ = np.concatenate([
        syndata_[continuous_features].values,
        scaler.transform(syndata_[categorical_features]).toarray()
    ], axis=1)
    
    n_steps = 30
    alphas = np.linspace(0, 1, n_steps) 
    
    emb_center = np.mean(train_, axis=0) # true embedding center
    synth_center = np.mean(syndata_, axis=0) # synthetic embedding center
    
    # L2 distance from true to embedding center
    dist = np.sqrt(((train_ - emb_center) ** 2).sum(axis=1)) 
    # Ball with quantiles of radii 
    # = approximation of the subset that supports a probability mass of alpha
    Radii = np.quantile(dist, alphas) 
    
    # L2 distance from synthetic to embedding center
    synth_to_center = np.sqrt(((syndata_ - emb_center) ** 2).sum(axis=1))
    
    nbrs_real = NearestNeighbors(n_neighbors=2, n_jobs=-1, p=2).fit(train_)
    real_to_real, _ = nbrs_real.kneighbors(train_) # distance to neighbors

    nbrs_synth = NearestNeighbors(n_neighbors=1, n_jobs=-1, p=2).fit(syndata_)
    # (distance to neighbors, indices of closest synthetic data point to real data)
    real_to_synth, real_to_synth_args = nbrs_synth.kneighbors(train_) 
    
    # Let us find closest real point to any real point, excluding itself (therefore, 1 instead of 0)
    real_to_real = real_to_real[:, 1].squeeze()
    real_to_synth = real_to_synth.squeeze()
    real_to_synth_args = real_to_synth_args.squeeze()
    
    # closest synthetic data points
    # = approximation of true data points using synthetic data points
    real_synth_closest = syndata_[real_to_synth_args] 
    real_synth_closest_d = np.sqrt(((real_synth_closest - synth_center) ** 2).sum(axis=1)) 
    # Ball with quantiles of Radii
    # = approximation of the subset that supports a probability mass of beta
    closest_synth_Radii = np.quantile(real_synth_closest_d, alphas)
    
    alpha_precision_curve = []
    beta_recall_curve = []
    for k in range(len(Radii)):
        alpha_precision = np.mean(
            synth_to_center <= Radii[k]
        )
        beta_recall = np.mean(
            (real_synth_closest_d <= closest_synth_Radii[k]) * (real_to_synth <= real_to_real)
        )
        alpha_precision_curve.append(alpha_precision)
        beta_recall_curve.append(beta_recall)
    
    # Riemann integral
    delta_precision_alpha = 1 - 2 * np.abs(alphas - np.array(alpha_precision_curve)).sum() * (alphas[1] - alphas[0])
    delta_beta_recall = 1 - 2 * np.abs(alphas - np.array(beta_recall_curve)).sum() * (alphas[1] - alphas[0])
    return delta_precision_alpha, delta_beta_recall
#%%


from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()
    
with open('requirements.txt', encoding='utf-8') as f:
    install_requires = f.read().splitlines()

setup(
    name='synthetic-eval',
    version='0.0.9',
    author='Seunghwan An',
    author_email='dpeltms79@gmail.com',
    description='Package for Evaluation of Synthetic Tabular Data Quality',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/an-seunghwan/synthetic_eval',
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
    ],
    python_requires='>=3.11',
    install_requires=install_requires
)


# %%
from collections import namedtuple
from . import metric_stat, metric_MLu, metric_privacy

import warnings
warnings.filterwarnings("ignore", "use_inf_as_na")

Metrics = namedtuple(
    "Metrics",
    [
        "KL",
        "GoF",
        "MMD",
        "CW",
        "alpha_precision", 
        "beta_recall",
        "base_cls", 
        "syn_cls",
        "model_selection", 
        "feature_selection",
        "Kanon_base",
        "Kanon_syn",
        "KMap",
        "DCR_RS",
        "DCR_RR",
        "DCR_SS",
        "AD",
    ]
)
#%%
def evaluate(syndata, train, test, target, continuous_features, categorical_features, device="cpu"):
    
    print("\n1. Statistical Fidelity\n")
    
    print("(marginal) KL-Divergence...\n")
    KL = metric_stat.KLDivergence(
        train, syndata, continuous_features)
    
    print("(marginal) Goodness Of Fit...\n")
    GoF = metric_stat.GoodnessOfFit(
        train, syndata, continuous_features)
    
    print("(joint) MMD...\n")
    MMD = metric_stat.MaximumMeanDiscrepancy(
        train, syndata, continuous_features, categorical_features)
    
    print("(joint) Cramer-Wold Distance...\n")
    CW = metric_stat.CramerWoldDistance(
        train, syndata, continuous_features, categorical_features, device)
    
    print("(joint) alpha-precision, beta-recall...\n")
    alpha_precision, beta_recall = metric_stat.naive_alpha_precision_beta_recall(
        train, syndata, continuous_features, categorical_features)
    
    print("\n2. Machine Learning Utility\n")
    
    print("Classification downstream task...\n")
    base_cls, syn_cls, model_selection, feature_selection = metric_MLu.MLu_cls(
        train, test, syndata, target, continuous_features, categorical_features)
    
    print("\n3. Privacy Preservability\n")
    
    print("K-anonimity...\n")
    Kanon_base, Kanon_syn = metric_privacy.kAnonymization(
        train, syndata, continuous_features)
    
    print("K-Map...\n")
    KMap = metric_privacy.kMap(
        train, syndata, continuous_features)
    
    print("Distance to Closest Record...\n")
    DCR_RS, DCR_RR, DCR_SS = metric_privacy.DCR_metric(
        train, syndata, continuous_features, categorical_features)
    
    print("Attribute Disclosure...\n")
    AD = metric_privacy.AttributeDisclosure(
        train, syndata, continuous_features, categorical_features)
    
    return Metrics(
        KL, GoF, MMD, CW, alpha_precision, beta_recall,
        base_cls, syn_cls, model_selection, feature_selection,
        Kanon_base, Kanon_syn, KMap, DCR_RS, DCR_RR, DCR_SS, AD
    )
#%%